////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File:        demo.c
// Function:    Sensoray DVR communication protocol, example utility functions and usage
// Copyright:   Copyright (c) Sensoray Co., Inc 2023
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <libgen.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <locale.h>


#include "dvr_simple.h"

static char *get_err_string(int err)
{
    switch(err) {
#define CASE_STR(x) case x: return #x
    CASE_STR(DVR_ERR_INVALID_OPCODE);
    CASE_STR(DVR_ERR_ILLEGAL_ARGVAL);
    CASE_STR(DVR_ERR_MISSING_ARG);
    CASE_STR(DVR_ERR_FUNC_PROHIBITED);
    CASE_STR(DVR_ERR_FILE_NOT_FOUND);
// DVR/API
    CASE_STR(DVR_ERR_MALLOC);
// API and transaction:
    CASE_STR(DVR_ERR_OPEN);
    CASE_STR(DVR_ERR_CLOSE);
    CASE_STR(DVR_ERR_SOCKCREATE);
    CASE_STR(DVR_ERR_SOCKBIND);
    CASE_STR(DVR_ERR_RXTHREAD);
    CASE_STR(DVR_ERR_TXTHREAD);
    CASE_STR(DVR_ERR_MUTEXCREATE);
    CASE_STR(DVR_ERR_LOCKCLOSED);
    CASE_STR(DVR_ERR_SOCKCLOSED);
    CASE_STR(DVR_ERR_OVERSIZE);
    CASE_STR(DVR_ERR_PACKETSEND);
    CASE_STR(DVR_ERR_PACKETREAD);
    CASE_STR(DVR_ERR_RUNTPACKET);
    CASE_STR(DVR_ERR_MISSINGDELIM);
    CASE_STR(DVR_ERR_CHECKSUM);
    CASE_STR(DVR_ERR_MSGTYPE);
    CASE_STR(DVR_ERR_DEPLETED);
    CASE_STR(DVR_ERR_STRINGBUF);
    CASE_STR(DVR_ERR_TIMEOUT);
    CASE_STR(DVR_ERR_EVENTTYPE);
    CASE_STR(DVR_ERR_STORAGE);
    CASE_STR(DVR_ERR_VIDLOCK);
// Menu system
    CASE_STR(DVR_ERR_MENUPANE_NOTFOUND);
    CASE_STR(DVR_ERR_MENUITEM_NOTFOUND);
    CASE_STR(DVR_ERR_MENUALREADYOPEN);
    CASE_STR(DVR_ERR_MENUCLOSED);
    }
    return "";
}

static char *get_status_string(u32 st)
{
    static char s[200] = "";

    strcpy(s, "");
    if (st & DVR_STAT_RECORDING) strcat(s, "REC ");
    if (st & DVR_STAT_PLAYING) strcat(s, "PLAY ");
    if (st & DVR_STAT_VIDEO_LOCK) strcat(s, "VIDLOCK ");
    if (st & DVR_STAT_KBD_ATTACHED) strcat(s, "KBD ");
    if (st & DVR_STAT_MEM_ATTACHED) strcat(s, "MEM ");
    if (st & DVR_STAT_WIFI_ENABLED) strcat(s, "WIFI ");
    if (st & DVR_STAT_RECORD_PAUSED) strcat(s, "RECPAUSE ");
    if (st & DVR_STAT_PLAY_PAUSED) strcat(s, "PLAYPAUSE ");
    if (st & DVR_STAT_MENU_OPEN) strcat(s, "MENU ");
    if (st & DVR_STAT_VIEW_SNAPSHOT) strcat(s, "SNAPSHOT ");
    if (st & DVR_STAT_WRITE_ERROR) strcat(s, "ERRWRITE ");

    return s;
}

static void print_error(int errcode)
{
    fprintf(stderr, "error %s: last status=%s\n",
        get_err_string(errcode),
        get_status_string(dvr_get_last_status()));
}





static int read_file_from_dvr(DVR_STORAGE_DEVID devid, const char *filename)
{
    FILE *f;
    u8 buf[MAX_CHUNKSIZE];
    int rc;
    int errcode = dvr_open_file(devid, DVR_FILE_MODE_READ, filename);
    if (errcode != DVR_ERR_OK) {
        fprintf(stderr, "Open file %s for reading: %s\n", filename, get_err_string(errcode));
        return errcode;
    }

    {
        char *name = strdup(filename);
        if (!name) {
            dvr_close_file();
            return DVR_ERR_MALLOC;
        }
        f = fopen(basename(name), "wb");
        free(name);
    }

    if (!f) {
        fprintf(stderr, "Failed to open destination file: %s\n", filename);
        dvr_close_file();
        return DVR_ERR_FILE_NOT_FOUND;
    }

    while (!ferror(f)) {
        u8 bytesread;
        errcode = dvr_read_file_chunk(buf, MAX_CHUNKSIZE, &bytesread);
        if (errcode != DVR_ERR_OK) {
            fprintf(stderr, "DVR Read file chunk failed: %s\n", get_err_string(errcode));
            break;
        }
        if (bytesread == 0) {
            break; // all done
        }
        errcode = fwrite(buf, 1, bytesread, f);
        if (errcode < 0) {
            fprintf(stderr, "Write file chunk failed: %s\n", strerror(errno));
            break;
        }
    }
    if (dvr_close_file() != DVR_ERR_OK) {
        fprintf(stderr, "Error closing file %s for writing: \n", filename);
    }
    fclose(f);
    return errcode;
}

static int write_file_to_dvr(DVR_STORAGE_DEVID devid, const char *filename)
{
    FILE *f;
    u8 buf[MAX_CHUNKSIZE];
    int rc;
    int errcode = dvr_open_file(devid, DVR_FILE_MODE_WRITE, filename);
    if (errcode != DVR_ERR_OK) {
        fprintf(stderr, "Open file %s for writing: %s\n", filename, get_err_string(errcode));
        return errcode;
    }

    {
        char *name = strdup(filename);
        if (!name) {
            dvr_close_file();
            return DVR_ERR_MALLOC;
        }
        f = fopen(basename(name), "rb");
        free(name);
    }

    if (!f) {
        fprintf(stderr, "Failed to open source file: %s\n", filename);
        dvr_close_file();
        return DVR_ERR_FILE_NOT_FOUND;
    }

    while (!feof(f)) {
        int bytesread = fread(buf, 1, MAX_CHUNKSIZE, f);
        if (bytesread < 0) {
            fprintf(stderr, "Read file chunk failed: %s\n", strerror(errno));
            break;
        }
        if (bytesread == 0) {
            break; // all done
        }
        errcode = dvr_write_file_chunk(buf, bytesread);
        if (errcode != DVR_ERR_OK) {
            fprintf(stderr, "DVR Write file chunk failed: %s\n", get_err_string(errcode));
            break;
        }
    }
    if (dvr_close_file() != DVR_ERR_OK) {
        fprintf(stderr, "Error closing file %s for reading: \n", filename);
    }
    fclose(f);
    return errcode;
}


static int read_buf_from_dvr(DVR_STORAGE_DEVID devid, const char *filename, u8 **pbuf, int *plen)
{
    u8 *buf = NULL;
    int buflen = 0;
    int rc;
    int errcode = dvr_open_file(devid, DVR_FILE_MODE_READ, filename);
    if (errcode != DVR_ERR_OK) {
        fprintf(stderr, "Open file %s for reading: %s\n", filename, get_err_string(errcode));
        return errcode;
    }

    while (buflen > 0) {
        u8 bytesread;
        buf = realloc(buf, buflen + MAX_CHUNKSIZE);
        if (!buf) {
            // fixme this will leak previous buf
            errcode = DVR_ERR_MALLOC;
            break;
        }

        errcode = dvr_read_file_chunk(buf + buflen, MAX_CHUNKSIZE, &bytesread);
        if (errcode != DVR_ERR_OK) {
            fprintf(stderr, "DVR Read file chunk failed: %s\n", get_err_string(errcode));
            break;
        }
        if (bytesread == 0) {
            break; // all done
        }
        buflen += bytesread;
    }

    if (dvr_close_file() != DVR_ERR_OK) {
        fprintf(stderr, "Error closing file %s for writing: \n", filename);
    }
    if (errcode == DVR_ERR_OK) {
        *pbuf = buf;
        *plen = buflen;
    }
    return errcode;
}

static int write_buf_to_dvr(DVR_STORAGE_DEVID devid, const char *filename, u8 *buf, int buflen)
{
    int rc;
    int errcode = dvr_open_file(devid, DVR_FILE_MODE_WRITE, filename);
    if (errcode != DVR_ERR_OK) {
        fprintf(stderr, "Open file %s for writing: %s\n", filename, get_err_string(errcode));
        return errcode;
    }

    while (buflen > 0) {
        int bytesread = buflen < MAX_CHUNKSIZE ? buflen : MAX_CHUNKSIZE;
        errcode = dvr_write_file_chunk(buf, bytesread);
        if (errcode != DVR_ERR_OK) {
            fprintf(stderr, "DVR Write file chunk failed: %s\n", get_err_string(errcode));
            break;
        }
        buf += bytesread;
        buflen -= bytesread;
    }
    if (dvr_close_file() != DVR_ERR_OK) {
        fprintf(stderr, "Error closing file %s for reading: \n", filename);
    }
    return errcode;
}



// which: 0=snapshot 1=record
static int metadata_file_to_dvr(u8 which, const char *filename)
{
    FILE *f;
    u8 buf[MAX_CHUNKSIZE];
    int errcode;

    f = fopen(filename, "rb");
    if (!f) {
        fprintf(stderr, "Failed to open metadata file: %s\n", filename);
        return DVR_ERR_FILE_NOT_FOUND;
    }

    // Send a zero-length buffer to clear current metadata
    errcode = (which ? dvr_set_record_metadata : dvr_set_snapshot_metadata)(NULL, 0);

    while (!feof(f)) {
        int bytesread = fread(buf, 1, MAX_CHUNKSIZE, f);
        if (bytesread <= 0) {
            if (bytesread < 0)
                fprintf(stderr, "Read metadata chunk failed: %s\n", strerror(errno));
            break;
        }
        // Send each chunk
        errcode = (which ? dvr_set_record_metadata : dvr_set_snapshot_metadata)(buf, bytesread);
        if (errcode != DVR_ERR_OK) {
            fprintf(stderr, "DVR Write metadata chunk failed: %s\n", get_err_string(errcode));
            break;
        }
    }
    fclose(f);
    return errcode;
}



// execute x and print the error if there was one
#define TEST(x)  \
do {  \
    int errcode = (x); \
    if (errcode != DVR_ERR_OK) { \
        fprintf(stderr, "%s: err=%d (%s) [%s]\n", #x, errcode, \
            get_err_string(errcode), \
            get_status_string(dvr_get_last_status())); \
    } \
} while (0)




static int tokenize(char *input, char **cmdv, int limit)
{
    int n = 0;
    while (n < limit) {
        while (*input && isblank(*input)) {
            ++input;
        }
        if (*input == '\0')
            break;
        *cmdv++ = input;
        n++;
        while (*input && !isblank(*input)) {
            ++input;
        }
        if (*input == '\0')
            break;
        *input++ = '\0';
    }
    return n;
}

// returns 1 if 'a' case-insensitive matches b or c (a, b or c may be NULL)
// otherwise 0
static int eq(const char *a, const char *b, const char *c)
{
    if (!a)
        return 0;
    if (b && strcasecmp(a, b) == 0)
        return 1;
    if (c && strcasecmp(a, c) == 0)
        return 1;
    return 0;
}


// get device index (from cur_devid if valid, otherwise from s and increment arg_idx)
static int get_devid(const char *s, int cur_devid, int *arg_idx)
{
    if (cur_devid != -1) return cur_devid;
    (*arg_idx)++;
    return eq(s, "int", "i") ? INTERNAL :
           eq(s, "ext", "e") ? EXTERNAL  :
           eq(s, "both", "b") ? BOTH :
           eq(s, "mmc", "m") ? MMC :
           -1;
}
// not all systems have all storage device names
const char *storage = "int/ext/mmc";
const char *storage_both = "int/ext/both/mmc";


////////////////////////////////////////////////////////////////////////////
// Example usage: Send overlay text to DVR using in-place encode/decode.

int main(int unused)
{
    int errcode;
    u32 status;     // dvr status flags
    int quit = 0;
    char *line = NULL;
    u32 api_version = 0;
    int max_osd_regions = 16;  // API Ver 1.1=16 1.2=17
    int cur_devid = -1;
    char *cmdq[5] = {"hw","a", NULL};

    setlocale(LC_NUMERIC, ""); // for printing numbers with commas, using printf("%'d"...

    errcode = open_com_port();
    if (errcode != DVR_ERR_OK) {
        fprintf(stderr, "%s: err=%d (%s)\n", "open_com_port", errcode,
            get_err_string(errcode));
	return 0;
    }

    while (!quit) {
        int cmdc;
        char *cmdv[32] = {};
        int idx = 0;

        if (cmdq[0]) {
            // copy startup commands from the command queue
            cmdv[0] = cmdq[0];
            cmdv[1] = NULL;
            // move up the command queue and fill empty with NULL
            memmove(cmdq, cmdq+1, sizeof(cmdq)-sizeof(cmdq[0]));
            cmdq[sizeof(cmdq)/sizeof(cmdq[0])-1] = NULL;

        } else {
            line = readline("> ");
            if (!line) break;

            cmdc = tokenize(line, cmdv, 32);
            if (!cmdv[0]) continue;
        }

        if (eq(cmdv[0], "q", "quit")) {
            quit = 1;
        } else if (eq(cmdv[0], "h", "help")) {
            printf("Help:\n"
                    " q | quit - exit this program\n"
                    " a | api - get the API version\n"
                    "hw | hardware - get the hardware version\n"
                    "st | status - get the DVR status\n"
                    "lp | lowpower [mode] - set low power mode: 0=normal 1=low power 2=deep sleep\n"
                    //"dm | dispmode [mode] - get/set the display mode\n"
                    "dt | datetime [now] - get/set the date and time\n"
                    "res [width] [height] - set the capture resolution (2960 only)\n"
                    "----- Recording ---------------------\n"
                    "note: [storage] is one of\n"
                    "     (model 4011,4023) int(i) / ext(e) / both(b)\n"
                    "     (model 4067,4031,2960tc) mmc(m)\n"
                    " r | rec [storage] [bitrate] - start recording MP4\n"
                    "rt | rects [storage] [bitrate] - start recording TS\n"
                    "rs | recstop - stop recording\n"
                    "rp | recpause - pause recording\n"
                    "rr | recresume - resume recording\n"
                    "rf | recfile [filename] - set recording filename\n"
                    "rm | recmeta [filename] - set recording metadata\n"
                    "rem | recenmeta [boolean] - enable recording metadata\n"
                    "----- Playback ----------------------\n"
                    " p | play [storage] [filename] - start playback\n"
                    "ps | playstop - stop playback\n"
                    "pp | playpause - pause playback\n"
                    "pr | playresume - resume playback\n"
                    "sk | playskip [n] - playback skip forward by n seconds\n"
                    "                    (or backward if n is negative)\n"
                    "sp | playspeed [normal/forward/reverse] - get/set playback speed\n"
                    "----- Snapshots ---------------------\n"
                    " s | snap [storage] [quality] - grab snapshot\n"
                    "sv | snapview - view snapshot\n"
                    "ss | snapstop - stop viewing snapshot\n"
                    "sf | snapfile [filename] - set snapshot filename\n"
                    "sm | snapmeta [filename] - set snapshot metadata\n"
                    "sem | snapenmeta [boolean] - enable snapshot metadata\n"
                    "----- File system ---------------------\n"
                    " dir | files [storage] - list files\n"
                    " del | delete [storage] [filename] - delete file\n"
                    " ren | rename [storage] [old] [new] - rename file\n"
                    "  fs | filesysinfo [storage] - show free/total space\n"
                    "send | sendfile [storage] [filename] - send file to DVR\n"
                    "recv | recvfile [storage] [filename] - receive file from DVR\n"
                    "----- Overlays ----------------------\n"
                    " o | osdtext [id] [string] - set osd text, id: overlay region 0 to 15\n"
                    "op | osdpos [id] [x] [y] - set osd position\n"
                    "oe | osdenable [id] [0/1/2/3] - get/set osd enabled and size\n"
                    "     0=off 1=8x14 2=16x16 3=19x32(HD only) 3=38x64(HD only)\n"
                    "bg | osdbg [id] [bgmode] - get/set osd background\n"
                    "   bgmode: 0=white/black 1=white/transparent\n"
                    "           4=black/transparent 5=black/white 6=white/blue(HD only)\n"
                    "cf | osdclockfmt [datefmt] [timefmt] - get/set osd clock format\n"
                    "   datefmt: 0=MM-DD-YYYY 1=DD-MM-YYYY 2=MM-DD-YY 3=DD-MM-YY\n"
                    "            4=MMM-DD-YYYY 5=DD-MMM-YYYY 6=MMM-DD-YY 7=DD-MMM-YY\n"
                    "            8=MMM DD, YYYY 9=DDMMMYYYY 10=MMM DD, YY 11=DDMMMYY\n"
                    "            12=YYYYMMDD 13=YYYYMMMDD 14=YYMMDD 15=YYMMMDD\n"
                    "   timefmt: 0=HH:MM:SS 1=HH:MM:SS.S 2=HH:MM:SS.SS 3=HH:MM:SS.SSS\n"
                    "            4=HH:MM 5=HH:MM AM/PM 6=HH:MM a/p\n"
                    "----- Video adjust/config -----------\n"
                    "va | vidadj [param] [value] - get/set video adjustment\n"
                    "     param: brightness, autobright, contrast, hue,\n"
                    "           saturation, chromafilter, combfilter, lumafilter\n"
                    "vc | vidcfg [format] [deint] [decimate] - get/set video config\n"
                    "----- Audio -------------------------\n"
                    "as   | audiosettings - get audio settings\n"
                    "inp  | audioinput [mic/line/digital] - get/set audio input\n"
                    "vol  | audiovolume [left/right/both] [vol] - get/set audio volume\n"
                    "mute | audiomute [left/right/both] [on/off] - get/set audio mute\n"
                    "agc  | audioautovol [left/right/both] [on/off] - get/set audio auto volume\n"
                    "----- Incremental encoder -----------\n"
                    "e  | encoder [chan] [count] - get/set encoder count\n"
                    "sc | encoderscalar [chan] [scale] - get/set encoder scalar\n"
                    "ob | encoderoffboard [chan] [on/off] - get/set off board encoder\n"
                    "es | encodersettings [chan] - get encoder settings\n"

                    );

        //=========== SYSTEM commands ====================

        } else if (eq(cmdv[0], "api", "a")) {
            TEST( dvr_get_api_version(&api_version) );
            printf("The DVR API version is %d.%d.%d\n",
                    (api_version >> 24) & 0xff,
                    (api_version >> 16) & 0xff,
                    (api_version & 0xffff));
            if (api_version < 0x01010000) {
                printf("DVR API version needs to be at least 1.1\n");
                return 0;
            }
            if (api_version >= 0x01020000) {
                max_osd_regions = 17;
            }

        } else if (eq(cmdv[0], "hardware", "hw")) {
            u32 version = 0;
            char model[MAX_FILENAME_SIZE] = "";
            TEST( dvr_get_hardware_info(&version, model) );
            printf("version=%x model=%s\n", version, model);
            if (cmdq[0] && eq(model, "S2960", NULL)) {
                cmdq[1] = "mmc";
            }

        } else if (eq(cmdv[0], "status", "st")) {
            u32 status = 0;
            TEST( dvr_get_dvr_status(&status) );
            printf("%s\n", get_status_string(status));

        } else if (eq(cmdv[0], "lowpower", "lp")) {
            if (!cmdv[1]) {
                printf("low power mode required: 0=normal 1=low power 2=deep sleep\n");
            } else {
                TEST( dvr_set_low_power(atoi(cmdv[1])) );
            }


        } else if (eq(cmdv[0], "datetime", "dt")) {
            time_t t;
            u32 clock;

            if (eq(cmdv[1], "now", NULL)) {
                struct timeval tv;
                struct timezone tz;
                gettimeofday(&tv, &tz);
                t = tv.tv_sec;
                clock = timegm(localtime(&t));  // adjust for local timezone

                printf("Set clock on the DVR to %s\n", asctime(localtime(&t)));

                TEST( dvr_set_datetime(clock) );

            } else {
                TEST( dvr_get_datetime(&clock) );

                t = clock;
                printf("The clock on the DVR is %s\n", asctime(gmtime(&t)));
            }

        } else if (eq(cmdv[0], "res", NULL)) {
            if (!cmdv[1] || !cmdv[2]) {
                printf("Width and Height required\n");
            } else {
                TEST( dvr_set_capture_res(atoi(cmdv[1]), atoi(cmdv[2])) );
            }

        //=========== RECORDING commands ====================

        } else if (eq(cmdv[0], "rec", "r") || eq(cmdv[0], "rects", "rt")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1) {
                printf("%s required\n", storage_both);
            } else {
                int fmt = (eq(cmdv[0], "rec", "r")) ? RECFMT_MP4 : RECFMT_TS;
                char filename[MAX_FILENAME_SIZE] = "";
                int err = dvr_record_start(devid, fmt, cmdv[idx+1] ? atoi(cmdv[idx+1]) : 0, filename);

                if (err == DVR_ERR_OK) {
                    printf("recording to %s\n", filename);
                } else {
                    print_error(err);
                }
            }

        } else if (eq(cmdv[0], "recstop", "rs")) {
            TEST( dvr_record_stop() );

        } else if (eq(cmdv[0], "recpause", "rp")) {
            TEST( dvr_record_pause() );

        } else if (eq(cmdv[0], "recresume", "rr")) {
            TEST( dvr_record_resume() );

        } else if (eq(cmdv[0], "recfile", "rf")) {
            char *filename = cmdv[1];

            if (!filename) {
                printf("filename required\n");
            } else {
                TEST( dvr_set_record_filename(filename) );
            }

        } else if (eq(cmdv[0], "recmeta", "rm")) {
            char *filename = cmdv[1];

            if (!filename) {
                printf("filename required\n");
            } else {
                TEST( metadata_file_to_dvr(1, filename) );
            }

        } else if (eq(cmdv[0], "recenmeta", "rem")) {
            if (eq(cmdv[1], "0", "1")) {
                TEST( dvr_enable_record_metadata(cmdv[1][0]-'0') );
            }


        //=========== PLAYBACK commands ====================

        } else if (eq(cmdv[0], "play", "p")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);
            char *filename = cmdv[idx+1];

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else if (!filename) {
                printf("filename required\n");
            } else {
                TEST( dvr_play_start(devid, filename) );
            }

        } else if (eq(cmdv[0], "playstop", "ps")) {
            TEST( dvr_play_stop() );

        } else if (eq(cmdv[0], "playpause", "pp")) {
            TEST( dvr_play_pause() );

        } else if (eq(cmdv[0], "playresume", "pr")) {
            TEST( dvr_play_resume() );

        } else if (eq(cmdv[0], "playspeed", "ps")) {
            if (cmdv[1]) {
                if (eq(cmdv[1], "normal", "n")) {
                    TEST( dvr_set_play_speed(PLAY_NORMAL) );
                } else if (eq(cmdv[1], "forward", "ff")) {
                    TEST( dvr_set_play_speed(PLAY_FAST_FORWARD) );
                } else if (eq(cmdv[1], "reverse", "rw")) {
                    TEST( dvr_set_play_speed(PLAY_FAST_REVERSE) );
                } else {
                    printf("invalid play speed\n");
                }
            } else {
                u8 speed = 0;
                TEST( dvr_get_play_settings(&speed) );
                printf("play speed is %s\n",
                        speed == PLAY_NORMAL ? "normal" :
                        speed == PLAY_FAST_FORWARD ? "fast forward" :
                        speed == PLAY_FAST_REVERSE ? "fast reverse" :
                        "invalid");
            }

        } else if (eq(cmdv[0], "playskip", "sk")) {
            if (!cmdv[1]) {
                printf("skip amount is required\n");
            } else {
                TEST( dvr_play_skip(atoi(cmdv[1])) );
            }

        //=========== SNAPSHOT commands ====================

        } else if (eq(cmdv[0], "snap", "s")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1) {
                printf("%s required\n", storage_both);
            } else {
                char filename[MAX_FILENAME_SIZE] = "";
                int err = dvr_grab_snapshot(devid, cmdv[idx+1] ? atoi(cmdv[idx+1]) : 0, filename);

                if (err == DVR_ERR_OK) {
                    printf("recording to %s\n", filename);
                } else {
                    print_error(err);
                }
            }

        } else if (eq(cmdv[0], "snapview", "sv")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);
            char *filename = cmdv[idx+1];

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else if (!filename) {
                printf("filename required\n");
            } else {
                TEST( dvr_view_snapshot(devid, filename) );
            }

        } else if (eq(cmdv[0], "ss", "snapstop")) {
            dvr_snapshot_stop();

        } else if (eq(cmdv[0], "snapfile", "sf")) {
            char *filename = cmdv[1];

            if (!filename) {
                printf("filename required\n");
            } else {
                TEST( dvr_set_snapshot_filename(filename) );
            }

        } else if (eq(cmdv[0], "snapmeta", "sm")) {
            char *filename = cmdv[1];

            if (!filename) {
                printf("filename required\n");
            } else {
                TEST( metadata_file_to_dvr(0, filename) );
            }

        } else if (eq(cmdv[0], "snapenmeta", "sem")) {
            if (eq(cmdv[1], "0", "1")) {
                TEST( dvr_enable_snapshot_metadata(cmdv[1][0]-'0') );
            }

        //=========== FILESYSTEM commands ====================

	} else if (eq(cmdv[0], "files", "dir")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else {
                u32 i, count = 0;
                TEST( dvr_get_filecount(devid, &count) );
                printf("filecount: %d  (devid=%d)\n", count, devid);
                for (i = 0; i < count; i++) {
                    char filename[MAX_FILENAME_SIZE];
                    u32 size;
                    //TEST( dvr_get_filename(devid, i, filename) );
                    TEST( dvr_get_filename_size(devid, i, filename, &size) );
                    printf("%-20s\t%d\n", filename, size);
                }
            }

        } else if (eq(cmdv[0], "delete", "del")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else if (!cmdv[2]) {
                printf("filename required\n");
            } else {
                TEST( dvr_delete_file(devid, cmdv[2]) );
            }

        } else if (eq(cmdv[0], "rename", "ren")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else if (!cmdv[idx+1]) {
                printf("old filename required\n");
            } else if (!cmdv[idx+2]) {
                printf("new filename required\n");
            } else {
                TEST( dvr_rename_file(devid, cmdv[idx+1], cmdv[idx+2]) );
            }
        } else if (eq(cmdv[0], "filesysinfo", "fs")) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else {
                u32 free_kb, total_kb = 0;
                TEST( dvr_get_filesystem_info(devid, &free_kb, &total_kb) );
                printf("%'d KB free / %'d KB total\n", free_kb, total_kb);
            }
        } else if (eq(cmdv[0], "send", NULL)) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else if (!cmdv[idx+1]) {
                printf("filename required\n");
            } else {
                TEST( write_file_to_dvr(devid, cmdv[idx+1]) );
            }

       	} else if (eq(cmdv[0], "recv", NULL)) {
            int devid = get_devid(cmdv[1], cur_devid, &idx);

            if (devid == -1 || devid == BOTH) {
                printf("%s required\n", storage);
            } else if (!cmdv[idx+1]) {
                printf("filename required\n");
            } else {
                TEST( read_file_from_dvr(devid, cmdv[idx+1]) );
            }

        //=========== OVERLAY commands ====================

        } else if (eq(cmdv[0], "o", "osdtext")) {
            int region = cmdv[1] ? atoi(cmdv[1]) : -1;

            if (region < 0 || region > max_osd_regions) {
                printf("region required\n");
            } else {
                int i;
                char *text = NULL;
                for (i = 2; i < 10 && cmdv[i]; i++) {
                    if (!text) {
                        text = strdup(cmdv[i]);
                    } else {
                        text = realloc(text, strlen(text) + 1 + strlen(cmdv[i]) + 1);
                        strcat(text, " ");
                        strcat(text, cmdv[i]);
                    }
                }
                char *tmp; // replace all "\\n" in text with single '\n' character
                while ((tmp = strstr(text, "\\n"))) {
                    *tmp = '\n';
                    // memmove must be used due to overlapping
                    memmove(tmp+1, tmp+2, strlen(tmp+2)+1);
                }
                TEST( dvr_set_osd_text(region, text) );
            }

        } else if (eq(cmdv[0], "op", "osdpos")) {
            int region = cmdv[1] ? atoi(cmdv[1]) : -1;
            int x = cmdv[2] ? atoi(cmdv[2]) : -1;
            int y = cmdv[3] ? atoi(cmdv[3]) : -1;

            if (region < 0 || region > max_osd_regions || x==-1 || y==-1) {
                printf("region, x, y required\n");
            } else {
                TEST( dvr_set_osd_position(region, x, y) );
            }

        } else if (eq(cmdv[0], "oe", "osdenable")) {
            int region = cmdv[1] ? atoi(cmdv[1]) : -1;
            int enable = cmdv[2] ? atoi(cmdv[2]) : -1;

            if (region < 0 || region > max_osd_regions) {
                printf("region, enable required\n");
            } else if (enable != -1) {
                TEST( dvr_set_osd_enable(region, enable) );
            } else {
                u8 enable;
                DVR_OSD_BGMODE bgmode;
                TEST( dvr_get_osd_settings(region, &enable, &bgmode) );
                printf("enable = %d\n", enable);
            }

        } else if (eq(cmdv[0], "bg", "osdbg")) {
            int region = cmdv[1] ? atoi(cmdv[1]) : -1;
            int bgmode = cmdv[2] ? atoi(cmdv[2]) : -1;

            if (region < 0 || region > max_osd_regions) {
                printf("region, enable required\n");
            } else if (bgmode != -1) {
                TEST( dvr_set_osd_background(region, bgmode) );
            } else {
                u8 enable;
                DVR_OSD_BGMODE bgmode;
                TEST( dvr_get_osd_settings(region, &enable, &bgmode) );
                printf("bgmode = %d\n", bgmode);
            }

        } else if (eq(cmdv[0], "cf", "osdclockfmt")) {
            if (!cmdv[1] && !cmdv[2]) {
                const char *dateformats[] = {
                    "MM-DD-YYYY","DD-MM-YYYY","MM-DD-YY","DD-MM-YY"
                    "MMM-DD-YYYY","DD-MMM-YYYY","MMM-DD-YY","DD-MMM-YY"
                    "MMM DD, YYYY","DDMMMYYYY","MMM DD, YY","DDMMMYn"
                    "YYYYMMDD","YYYYMMMDD","YYMMDD","YYMMMDD"
                };
                const char *timeformats[] = {
                    "HH:MM:SS","=HH:MM:SS.S","HH:MM:SS.SS","HH:MM:SS.SSS",
                    "HH:MM","HH:MM AM/PM","HH:MM a/p"
                };
                DVR_DATE_FORMAT df;
                DVR_TIME_FORMAT tf;

                TEST( dvr_get_osd_clock_format(&df, &tf) );

                printf("dateformat: %d (%s)\n", df, dateformats[df]);
                printf("timeformat: %d (%s)\n", tf, timeformats[tf]);
            } else if (cmdv[2]) {
                int df = atoi(cmdv[1]);
                int tf = atoi(cmdv[2]);

                TEST( dvr_set_osd_clock_format(df, tf) );

            } else {
                printf("datefmt and timefmt required\n");
            }


        //=========== VIDEO commands ====================

        } else if (eq(cmdv[0], "vidadj", "va")) {
            int param =
                eq(cmdv[1], "brightness","br") ? ADJ_BRIGHTNESS :
                eq(cmdv[1], "autobright","ab") ? ADJ_AUTOBRIGHT :
                eq(cmdv[1], "contrast", "co") ? ADJ_CONTRAST :
                eq(cmdv[1], "hue", "hu") ? ADJ_HUE :
                eq(cmdv[1], "saturation", "sat") ? ADJ_SATURATION :
                eq(cmdv[1], "chromafilter", "cf") ? ADJ_CHROMA_FILT :
                eq(cmdv[1], "lumafilter", "lf") ? ADJ_LUMA_FILT :
                eq(cmdv[1], "combfilter", "comb") ? ADJ_COMB_FILT :
                -1;
            const char *names[] = {
                "brightness",
                "autobright",
                "contrast",
                "hue",
                "saturation",
                "chromafilt",
                "lumafilt",
                "combfilt",
            };

            if (param == -1 && cmdv[1]) {
                printf("invalid param name\n");
            } else if (param == -1) {
                int id;
                for (id = 0; id <= ADJ_COMB_FILT; id++) {
                    u8 value;
                    TEST( dvr_get_video_adjust(id, &value) );
                    printf("%10s: %3d\n", names[id], id == ADJ_HUE ? (int)(signed char)value : value);
                }
            } else {
                if (cmdv[2]) {
                    TEST( dvr_set_video_adjust(param, atoi(cmdv[2])) );
                } else {
                    u8 value;
                    TEST( dvr_get_video_adjust(param, &value) );
                    printf("%s: %d\n", names[param], param == ADJ_HUE ? (int)(signed char)value : value);
                }
            }

        } else if (eq(cmdv[0], "vidcfg", "vc")) {
            u8 format, deint, decimate;
            if (!cmdv[1]) {
                TEST( dvr_get_video_config(&format, &deint, &decimate) );
                printf("Video config: format=%d deint=%d decimate=%d\n",
                    format, deint, decimate);
            } else if (!cmdv[2] || !cmdv[3]) {
                printf("format, deint, decimate required\n");
            } else {
                format = atoi(cmdv[1]);
                deint = atoi(cmdv[2]);
                decimate = atoi(cmdv[3]);

                TEST( dvr_set_video_config(format, deint, decimate) );
                printf("Video config: format=%d deint=%d decimate=%d\n",
                    format, deint, decimate);
            }

        //=========== AUDIO commands ====================

        }  else if (eq(cmdv[0], "audioinput", "inp") ||
                    eq(cmdv[0], "audiovolume", "vol") ||
                    eq(cmdv[0], "audiomute", "mute") ||
                    eq(cmdv[0], "audioautovol", "agc") ||
                    eq(cmdv[0], "audiosettings", "as")) {
            DVR_AUDIO_INPUT inp;
            u8 Lvol, Rvol, Lmute, Rmute, Lagc, Ragc;
            const char *audio_names[] = {"mic", "line", "digital"};
            const char *audio_chans[] = {"none", "left", "right", "both"};
            int chansel =
                    eq(cmdv[1], "left", "l") ? AUDIOCHAN_LEFT :
                    eq(cmdv[1], "right", "r") ? AUDIOCHAN_RIGHT :
                    eq(cmdv[1], "both", "b") ? AUDIOCHAN_BOTH :
                    -1;

            if (!cmdv[1] || eq(cmdv[0], "audiosettings", "as")) {
                TEST( dvr_get_audio_settings(&inp, &Lvol, &Rvol, &Lmute, &Rmute, &Lagc, &Ragc) );
                if (eq(cmdv[0], "audioinput", "inp") || eq(cmdv[0], "audiosettings", "as")) {
                    printf("Audio input: %s\n", audio_names[inp]);
                }
                if (eq(cmdv[0], "audiovolume", "vol") || eq(cmdv[0], "audiosettings", "as")) {
                    printf("Audio volume: left %d  right %d\n", Lvol, Rvol);
                }
                if (eq(cmdv[0], "audiomute", "mute") || eq(cmdv[0], "audiosettings", "as")) {
                    printf("Audio mute: %s\n", audio_chans[Lmute*2 + Rmute]);
                }
                if (eq(cmdv[0], "audioautovol", "agc") || eq(cmdv[0], "audiosettings", "as")) {
                    printf("Audio auto volume: %s\n", audio_chans[Lagc*2 + Ragc]);
                }
            } else if (eq(cmdv[0], "audiovolume", "vol")) {
                int volume = cmdv[2] ? atoi(cmdv[2]) : -1;
                if (chansel == -1 || volume == -1) {
                    printf("chansel and volume required\n");
                } else {
                    TEST( dvr_set_volume(chansel, volume) );
                }
            } else if (eq(cmdv[0], "audiomute", "mute")) {
                int mute = eq(cmdv[2], "on", "1") ? 1 :
                        eq(cmdv[2], "off", "0") ? 0 :
                        -1;
                if (chansel == -1 || mute == -1) {
                    printf("chansel and mute required\n");
                } else {
                    TEST( dvr_set_mute(chansel, mute) );
                }
            } else if (eq(cmdv[0], "audioautovol", "agc")) {
                int agc = eq(cmdv[2], "on", "1") ? 1 :
                        eq(cmdv[2], "off", "0") ? 0 :
                        -1;
                if (chansel == -1 || agc == -1) {
                    printf("chansel and agc required\n");
                } else {
                    TEST( dvr_set_autovol(chansel, agc) );
                }
            }

        //=========== INCR ENCODER commands ====================

        } else if (eq(cmdv[0], "encoder", "e")) {
            int chan = cmdv[1] ? atoi(cmdv[1]) : -1;
            if (chan == -1) {
                printf("chan is required\n");
            } else if (cmdv[2]) {
                u32 count = strtoul(cmdv[2], NULL, 10);
                TEST( dvr_set_encoder_counts(chan, count) );
            } else {
                u32 count = 0;
                TEST( dvr_get_encoder_counts(chan, &count) );
                printf("encoder %d count is %u\n", chan, count);
            }

        } else if (eq(cmdv[0], "encoderscalar", "sc")) {
            int chan = cmdv[1] ? atoi(cmdv[1]) : -1;
            if (chan == -1) {
                printf("chan is required\n");
            } else if (cmdv[2]) {
                double scale = strtod(cmdv[2], NULL);
                TEST( dvr_set_encoder_scalar(chan, scale) );
            } else {
                double scale;
                u8 use_offboard;
                TEST( dvr_get_encoder_settings(chan, &scale, &use_offboard) );
                printf("encoder %d scale is %f\n", chan, scale);
            }

        } else if (eq(cmdv[0], "encoderoffboard", "ob")) {
            int chan = cmdv[1] ? atoi(cmdv[1]) : -1;
            if (chan == -1) {
                printf("chan is required\n");
            } else if (cmdv[2]) {
                u8 enable = eq(cmdv[2], "on", "1") ? 1 :
                        eq(cmdv[2], "off", "0") ? 0 : -1;
                TEST( dvr_use_offboard_encoder(chan, enable) );
            } else {
                double scale;
                u8 use_offboard;
                TEST( dvr_get_encoder_settings(chan, &scale, &use_offboard) );
                printf("encoder %d offboard is %s\n", chan, use_offboard ? "enabled" : "disabled");
            }

        } else if (eq(cmdv[0], "encodersettings", "es")) {
            int chan = cmdv[1] ? atoi(cmdv[1]) : -1;
            if (chan == -1) {
                printf("chan is required\n");
            } else {
                double scale;
                u8 use_offboard;
                TEST( dvr_get_encoder_settings(chan, &scale, &use_offboard) );
                printf("encoder %d scale is %f, offboard is %s\n", chan,
                        scale, use_offboard ? "enabled" : "disabled");
            }
        } else if (eq(cmdv[0], "sleep", NULL)) {
            double f = atof(cmdv[1]);
            struct timespec ts;
            ts.tv_sec = f;
            ts.tv_nsec = (f - ts.tv_sec) * 1e9;
            nanosleep(&ts, NULL);

        //=========== device selection commands ====================

        } else if (eq(cmdv[0], "int", NULL)) {
            cur_devid = INTERNAL;
            printf("Current storage set to INTERNAL\n");
        } else if (eq(cmdv[0], "ext", NULL)) {
            cur_devid = EXTERNAL;
            printf("Current storage set to EXTERNAL\n");
        } else if (eq(cmdv[0], "mmc", NULL)) {
            cur_devid = MMC;
            printf("Current storage set to MMC\n");
        } else {
            printf("unrecognized commmand - use h for help\n");
        }

        free(line);
        line = NULL;
    }
    close_com_port();
}
